/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.18.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : entry.s
Purpose : Entry point and vector table for the FE310-G002.

Additional information:
  [1] The RISC-V Instruction Set Manual Volume II: Privileged Architecture
      ...
      3.1.7 Machine Trap-Vector Base-Address Register (mtvec)
        ...
        When vectored interrupts are enabled, interrupt cause 0, which corresponds to user-mode software interrupts,
        are vectored to the same location as synchronous exceptions. This ambiguity does not arise in practice,
        since user-mode software interrupts are either disabled or delegated to user mode.
        ...
*/

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#if (__riscv_xlen == 64)
  #define STORE     sd
  #define LOAD      ld
  #define REGBYTES   8
#else
  #define STORE     sw
  #define LOAD      lw
  #define REGBYTES   4
#endif

/*********************************************************************
*
*       MACROS
*
**********************************************************************
*/

/*********************************************************************
*
*      TRAP_ENTRY
*/
    .macro TRAP_ENTRY
        addi    sp, sp, -20 * REGBYTES
        //
        // Store caller save registers on the stack.
        // Callee save registers will be saved by the respective ISR function.
        //
        STORE   ra,  0 * REGBYTES(sp)
        STORE   t0,  1 * REGBYTES(sp)
        STORE   t1,  2 * REGBYTES(sp)
        STORE   t2,  3 * REGBYTES(sp)
        STORE   a0,  4 * REGBYTES(sp)
        STORE   a1,  5 * REGBYTES(sp)
        STORE   a2,  6 * REGBYTES(sp)
        STORE   a3,  7 * REGBYTES(sp)
        STORE   a4,  8 * REGBYTES(sp)
        STORE   a5,  9 * REGBYTES(sp)
        STORE   a6, 10 * REGBYTES(sp)
        STORE   a7, 11 * REGBYTES(sp)
        STORE   t3, 12 * REGBYTES(sp)
        STORE   t4, 13 * REGBYTES(sp)
        STORE   t5, 14 * REGBYTES(sp)
        STORE   t6, 15 * REGBYTES(sp)
        //
        // Store the mepc on the stack.
        // This is needed for nestable interrupts.
        //
        csrr    a1, mepc
        STORE   a1, 16 * REGBYTES(sp)
    .endm

/*********************************************************************
*
*      TRAP_EXIT
*/
    .macro TRAP_EXIT
        //
        // Load the mepc from the stack into the mepc register.
        // This is needed for nestable interrupts.
        //
        LOAD    a1, 16 * REGBYTES(sp)
        csrw    mepc, a1

        // Remain in M-mode after mret
        li      t0, 0x1800
        csrs    mstatus, t0
        //
        // Load caller save registers from stack and return.
        //
        LOAD    ra,  0 * REGBYTES(sp)
        LOAD    t0,  1 * REGBYTES(sp)
        LOAD    t1,  2 * REGBYTES(sp)
        LOAD    t2,  3 * REGBYTES(sp)
        LOAD    a0,  4 * REGBYTES(sp)
        LOAD    a1,  5 * REGBYTES(sp)
        LOAD    a2,  6 * REGBYTES(sp)
        LOAD    a3,  7 * REGBYTES(sp)
        LOAD    a4,  8 * REGBYTES(sp)
        LOAD    a5,  9 * REGBYTES(sp)
        LOAD    a6, 10 * REGBYTES(sp)
        LOAD    a7, 11 * REGBYTES(sp)
        LOAD    t3, 12 * REGBYTES(sp)
        LOAD    t4, 13 * REGBYTES(sp)
        LOAD    t5, 14 * REGBYTES(sp)
        LOAD    t6, 15 * REGBYTES(sp)
        addi    sp, sp, 20 * REGBYTES
        mret
    .endm

/*********************************************************************
*
*      WEAK_DECL
*/
    .macro WEAK_DECL Name
        .weak \Name
        .type \Name, function
\Name\():
    .endm

/*********************************************************************
*
*      STARTFUNC
*/
    .macro STARTFUNC FunctionName:req, SectionName=
        .ifb \SectionName                        // If parameter section name is blank
        .section .text.\FunctionName, "ax", %progbits
        .else
        .section \SectionName, "ax", %progbits
        .endif
        .type \FunctionName, "function"
        .global \FunctionName
        .balign 4
#if (!defined(__clang__))
        .func
#endif
\FunctionName:
    .endm

/*********************************************************************
*
*      ENDFUNC
*/
    .macro ENDFUNC FunctionName
#if (defined(__clang__))
\FunctionName\()_end:
        .size \FunctionName, \FunctionName\()_end-\FunctionName
#else
        .endfunc
#endif
    .endm

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       Reset_Handler
*/
        STARTFUNC Reset_Handler, .app_entry
//
// Jump to program start
//
        .option push
        .option norelax
        la      a0, _start
        jr      a0
        .option pop
        ENDFUNC Reset_Handler

/*********************************************************************
*
*       Vector table
*/
        STARTFUNC vtrap_entry, .vectors
        .balign 256
        .option push
        .option norvc
        j    USI_Handler   // This vector is ambiguously used for synchronous exceptions as well (see [1])
        j    SSI_Handler
        j    Undefined     // reserved, previously intended for Hypervisor mode
        j    MSI_Handler
        j    UTI_Handler
        j    STI_Handler
        j    Undefined     // reserved, previously intended for Hypervisor mode
        j    MTI_Handler
        j    UEI_Handler
        j    SEI_Handler
        j    Undefined     // reserved, previously intended for Hypervisor mode
        j    MEI_Handler
        j    Undefined     // reserved
        j    Undefined     // reserved
        j    Undefined     // reserved
        j    Undefined     // reserved
        .option pop
        ENDFUNC vtrap_entry

/*********************************************************************
*
*       USI_Handler
*/
        STARTFUNC USI_Handler
        TRAP_ENTRY
    //  call    ISR_U_Software  --- Synchronous exceptions and user mode software interrupts are ambiguously vectored to the same exception handler
        call    OS_TrapHandler  // Implemented in RTOSInit.c
        TRAP_EXIT
        ENDFUNC USI_Handler

/*********************************************************************
*
*       SSI_Handler
*/
        STARTFUNC SSI_Handler
        TRAP_ENTRY
        call    ISR_S_Software
        TRAP_EXIT
        ENDFUNC SSI_Handler

/*********************************************************************
*
*       MSI_Handler
*/
        STARTFUNC MSI_Handler
        TRAP_ENTRY
        call    ISR_M_Software  // Implemented in RTOSInit.c
        TRAP_EXIT
        ENDFUNC MSI_Handler

/*********************************************************************
*
*       UTI_Handler
*/
        STARTFUNC UTI_Handler
        TRAP_ENTRY
        call    ISR_U_Timer
        TRAP_EXIT
        ENDFUNC UTI_Handler

/*********************************************************************
*
*       STI_Handler
*/
        STARTFUNC STI_Handler
        TRAP_ENTRY
        call    ISR_S_Timer
        TRAP_EXIT
        ENDFUNC STI_Handler

/*********************************************************************
*
*       MTI_Handler
*/
        STARTFUNC MTI_Handler
        TRAP_ENTRY
        call    ISR_M_Timer  // Implemented in RTOSInit.c
        TRAP_EXIT
        ENDFUNC MTI_Handler

/*********************************************************************
*
*       UEI_Handler
*/
        STARTFUNC UEI_Handler
        TRAP_ENTRY
        call    ISR_U_External
        TRAP_EXIT
        ENDFUNC UEI_Handler

/*********************************************************************
*
*       SEI_Handler
*/
        STARTFUNC SEI_Handler
        TRAP_ENTRY
        call    ISR_S_External
        TRAP_EXIT
        ENDFUNC SEI_Handler

/*********************************************************************
*
*       MEI_Handler
*/
        STARTFUNC MEI_Handler
        TRAP_ENTRY
        call    ISR_M_External  // Implemented in RTOSInit.c
        TRAP_EXIT
        ENDFUNC MEI_Handler

/*** Weak high-level IRQ handler, defaulting to OS_ISR_Undefined(). Can be overwritten in "C" ***/
        WEAK_DECL ISR_U_Software
        WEAK_DECL ISR_S_Software
        WEAK_DECL ISR_U_Timer
        WEAK_DECL ISR_S_Timer
        WEAK_DECL ISR_U_External
        WEAK_DECL ISR_S_External
Undefined:  j       Undefined
        ret

/********************************************************************/

        .end

/*************************** End of file ****************************/
